# Functions for handling trees and coupling inference

import copy
import ete3
import networkx as nx
import newick
import numpy as np
import warnings
from cvxopt.solvers import qp as cvxopt_qp
from cvxopt import matrix as cvxopt_matrix
from numbers import Number

import lineageot.simulation as sim





###############################
# General inference functions #
###############################

def rate_estimator(barcode_array, time):
    """
    Estimates the mutation rate based on the number of unmutated
    barcodes remaining.
    """
    num_nonzeros = np.count_nonzero(barcode_array)
    num_zeros = barcode_array.size - num_nonzeros
    return -np.log(num_zeros/barcode_array.size)/time


def scaled_Hamming_distance(barcode1, barcode2):
    """
    Computes the distance between two barcodes, adjusted for

    (1) the number of sites where both cells were measured and

    (2) distance between two scars is twice the distance from
                            
        scarred to unscarred
    """
    shared_indices = (barcode1 >= 0) & (barcode2 >= 0)
    b1 = barcode1[shared_indices]

    # There may not be any sites where both were measured
    if len(b1) == 0:
        return np.nan
    b2 = barcode2[shared_indices]

    differences = b1 != b2
    double_scars = differences & (b1 != 0) & (b2 != 0)


    return (np.sum(differences) + np.sum(double_scars))/len(b1)

def barcode_distances(barcode_array):
    """
    Computes all pairwise lineage distances between barcodes
    """
    num_cells = barcode_array.shape[0]
    lineage_distances = np.zeros([num_cells, num_cells])

    for i in range(num_cells):
        barcode1 = barcode_array[i,:]
        for j in range(i):
            barcode2 = barcode_array[j]

            lineage_distances[i,j] = scaled_Hamming_distance(barcode1, barcode2)
            lineage_distances[j,i] = lineage_distances[i,j]
    return lineage_distances


def OT_cost(coupling, cost):
    return np.tensordot(coupling, cost, 2)


##########################
# Tree-related functions #
##########################

def list_tree_to_digraph(list_tree):
    """
    Converts a tree stored as nested lists to a networkx DiGraph

    Internal nodes are indexed by negative integers, leaves by nonnegative integers
    """
    next_internal_node = -1
    next_leaf_node = 0
    T, next_internal_node, next_leaf_node, subtree_root = recursive_list_tree_to_digraph(list_tree, 
                                                                                         next_internal_node, 
                                                                                         next_leaf_node)

    barcode_length = len(T.nodes[0]['cell'].barcode)
    T.add_node('root', cell=sim.Cell(np.nan, np.zeros(barcode_length)), time_to_parent = 0)
    T.add_edge('root', subtree_root, time=T.nodes[subtree_root]['time_to_parent'])

    return T

    



def recursive_list_tree_to_digraph(list_tree, next_internal_node, next_leaf_node):
    """
    Recursive helper function for list_tree_to_digraph

    Returns (current_tree, next_internal_node_label, root_of_current_tree)
    """


    if len(list_tree) == 1:
        assert(not isinstance(list_tree[0], list))
        T = nx.DiGraph()
        T.add_node(next_leaf_node, cell=list_tree[0][0], time_to_parent=list_tree[0][1])
        return T, next_internal_node, next_leaf_node + 1, next_leaf_node
    else:
        assert(len(list_tree) == 3) # assuming binary tree: list has two leaves and the cell at the current node
        left_subtree, next_internal_node, next_leaf_node, left_root = recursive_list_tree_to_digraph(list_tree[0],
                                                                                                     next_internal_node, 
                                                                                                     next_leaf_node)
        right_subtree, next_internal_node, next_leaf_node, right_root = recursive_list_tree_to_digraph(list_tree[1], 
                                                                                                       next_internal_node, 
                                                                                                       next_leaf_node)

        T = nx.compose(left_subtree, right_subtree)
        T.add_node(next_internal_node, cell=list_tree[2][0], time_to_parent=list_tree[2][1])
        T.add_edge(next_internal_node, left_root, time=T.nodes[left_root]['time_to_parent'])
        T.add_edge(next_internal_node, right_root, time=T.nodes[right_root]['time_to_parent'])

        return T, next_internal_node - 1, next_leaf_node, next_internal_node
        

















def annotate_tree(tree, mutation_rates, time_inference_method = 'independent', overwrite_times = False):
    """
    Adds barcodes and times to internal (ancestor) nodes so likelihoods can be computed

    Barcodes are inferred by putting minimizing the number of mutation events required,
    assuming a model with no back mutations and a known initial barcode
    """

    recursive_add_barcodes(tree, 'root')
    add_times(tree, mutation_rates, time_inference_method, overwrite = overwrite_times)
    add_times_to_edges(tree)
    return tree

def add_leaf_barcodes(tree, barcode_array):
    """
    Adds barcodes from barcode_array to the corresponding leaves of the tree
    """
    
    num_cells = barcode_array.shape[0]
    for cell in range(num_cells):
        if 'cell' in tree.nodes[cell]:
            tree.nodes[cell]['cell'].barcode = barcode_array[cell, :]
        else:
            tree.nodes[cell]['cell'] = sim.Cell(np.nan, barcode_array[cell, :])
    tree.nodes['root']['cell'] = sim.Cell(np.nan, np.zeros(barcode_array.shape[1]))
    return tree

def add_leaf_x(tree, x_array):
    """
    Adds expression vectors from x_array to the corresponding leaves of the tree
    """
    num_cells = x_array.shape[0]
    for cell in range(num_cells):
        if 'cell' in tree.nodes[cell]:
            tree.nodes[cell]['cell'].x = x_array[cell, :]
        else:
            tree.nodes[cell]['cell'] = sim.Cell(x_array[cell, :], np.nan)
    return tree

    


def recursive_add_barcodes(tree, current_node):
    """
    Fills in the barcodes for internal nodes for a tree whose leaves have barcodes

    Minimizes the number of mutation events that occur, assuming no backmutations
    and a known initial barcode
    """


    children = tree.successors(current_node)
    child_barcodes = []
    for child in children:
        recursive_add_barcodes(tree, child)
        child_barcodes.append(tree.nodes[child]['cell'].barcode)
    if 'cell' in tree.nodes[current_node]:
        return
    else:
        # if we've somehow ended up at a leaf with no cell, this will fail
        # likely this means that the tree was improperly annotated
        # (e.g., too few barcodes were added)
        current_barcode = child_barcodes[0].copy()
        if len(child_barcodes) == 1:
            warnings.warn("Some cell has only one descendant. Assuming no mutations occurred between them.")
        else:
            # assume that if a site is missing in one child it equals the
            # value in the other child
            current_barcode[child_barcodes[0] == -1] = child_barcodes[1][child_barcodes[0] == -1]
            # if the two children differ and neither is missing, set parent to unmutated
            current_barcode[(child_barcodes[0] != child_barcodes[1]) &
                            (child_barcodes[1] != -1)] = 0
        
        tree.nodes[current_node]['cell'] = sim.Cell(np.nan, current_barcode)
        return



def add_times(tree, mutation_rates, time_inference_method, overwrite = False):
    """
    Adds estimated division times/edge lengths to a tree

    The tree should already have all node barcodes estimated
    """
    for node in tree.nodes:
        if node == 'root':
            tree.nodes[node]['ml_time_to_parent'] = 0
        else:
            tree.nodes[node]['ml_time_to_parent'] = estimate_division_time(tree.nodes[node],
                                                                        tree.nodes[next(tree.predecessors(node))],
                                                                        mutation_rates
                                                                           )
    if time_inference_method == 'independent':
        # Infer times for edges entirely independently of each other,
        # ignoring constraints (like that leaves were sampled at the
        # same time)
        for node in tree.nodes:
            if (not ('time_to_parent' in tree.nodes[node])) or overwrite:
                tree.nodes[node]['time_to_parent'] = tree.nodes[node]['ml_time_to_parent']

    elif time_inference_method == 'least_squares':
        # Finds estimate of internal node times that minimizes the
        # squared error of the implied division times (relative to
        # the ml_time_to_parent estimate
        internal_nodes = get_internal_nodes(tree)
        leaves = get_leaves(tree)

        final_time = tree.nodes[leaves[0]]['time']
        # root should have time zero
        assert(tree.nodes[leaves[-1]]['time'] == 0)

        laplacian = nx.linalg.laplacian_matrix(nx.Graph(tree),
                                               nodelist = internal_nodes + leaves)
        incidence_matrix = nx.linalg.incidence_matrix(tree,
                                                      nodelist = internal_nodes + leaves,
                                                      oriented = True)
        
        edge_target_times = np.array([tree.nodes[e[1]]['ml_time_to_parent'] for e in tree.edges()])
        leaf_and_root_times = np.ones(len(leaves))*final_time
        leaf_and_root_times[-1] = 0

        b = (incidence_matrix[range(len(internal_nodes)), :] @ edge_target_times
             - laplacian[np.ix_(range(len(internal_nodes)),
                                range(len(internal_nodes), 
                                      len(internal_nodes)+len(leaves))
                                )] @ leaf_and_root_times)




        laplacian = np.array(laplacian.todense().astype(np.double))
        incidence_matrix = np.array(incidence_matrix.todense().astype(np.double))

        P = laplacian[np.ix_(range(len(internal_nodes)), range(len(internal_nodes)))]
        G = -incidence_matrix[range(len(internal_nodes)),:].T
        h = incidence_matrix[range(len(internal_nodes), len(internal_nodes) + len(leaves))].T @ leaf_and_root_times

        # In the least-squares internal_node_times may not be consistent with the ordering given by the tree
        # The quadratic program enforces the constraint that all edges have positive time
        internal_node_times = np.reshape(cvxopt_qp_from_numpy(P, -b, G, h), (len(internal_nodes)))
        add_node_times_from_dict(tree,
                                 'root',
                                 {node:time for node, time in zip(internal_nodes, internal_node_times)})
            
        add_division_times_from_vertex_times(tree)
                                               
    else:
        raise NotImplementedError(time_inference_method + ' is not an accepted time inference method')
    return


def estimate_division_time(child, parent, mutation_rates):
    """
    Estimates the lifetime of child, i.e. the time between when parent
    divided to make child and when child divided

    Input arguments are nodes in a lineage tree, i.e. dicts
    """

    #TODO: is there a better estimator?
    c_barcode = child['cell'].barcode
    p_barcode = parent['cell'].barcode

    available_entries = (c_barcode != -1) & (p_barcode == 0)
    differing_entries = ((c_barcode != p_barcode) &
                         available_entries
                         )
    
    #TODO: handle possible division by zero
    return np.sum(differing_entries)/np.sum(mutation_rates[available_entries])
        


def add_division_times_from_vertex_times(tree, current_node = 'root'):
    """
    Adds 'time_to_parent' variables to nodes, based on 'time' annotations
    """
    
    for child in tree.successors(current_node):
        tree.nodes[child]['time_to_parent'] = (tree.nodes[child]['time']
                                               - tree.nodes[current_node]['time']
                                               )
        add_division_times_from_vertex_times(tree, current_node = child)

    return
    

def add_times_to_edges(tree):
    """
    Labels each edge of tree with 'time' taken from
    'time_to_parent' of its endpoint
    """
    for (u, v) in tree.edges:
        t = tree.nodes[v]['time_to_parent']
        if t > 0:
            tree.edges[u,v]['time'] = t
        else:
            warnings.warn('Nonpositive time to parent encountered: t = '
                          + str(t)
                          + ' for edge (' + str(u) + ', ' + str(v) + ').')
            tree.edges[u,v]['time'] = np.finfo(float).eps
    return tree

def add_inverse_times_to_edges(tree):
    """
    Labels each edge of the tree with 'inverse time' equal
    to 1/edge['time']
    """
    for (u, v) in tree.edges:
        tree.edges[u,v]['inverse time'] = 1/tree.edges[u,v]['time']
    return tree


def add_leaf_times(tree, final_time):
    """
    Adds the known final time to all leaves
    and 0 as the root time
    """
    leaves = get_leaves(tree)
    for leaf in leaves:
        if leaf == 'root':
            tree.nodes[leaf]['time'] = 0
            tree.nodes[leaf]['time_to_parent'] = 0
        else:
            tree.nodes[leaf]['time'] = final_time
    return

def add_node_times_from_dict(tree, current_node, time_dict):
    """
    Adds times from time_dict to current_node and its descendants
    """
    if current_node in time_dict:
        # enforce children not having earlier times than parents
        tree.nodes()[current_node]['time'] = np.maximum(time_dict[current_node],
                                                        tree.nodes()[next(tree.predecessors(current_node))]['time'])
    for child in tree.successors(current_node):
        add_node_times_from_dict(tree, child, time_dict)

    return

def add_node_times_from_division_times(tree, current_node = 'root', overwrite = False):
    """
    Adds 'time' variable to all descendants of current_node based on the 'time_to_parent' variable
    """
    if current_node == 'root':
        if overwrite | (not 'time' in tree.nodes[current_node]):
            tree.nodes[current_node]['time'] = 0
    else:
        parent = next(tree.predecessors(current_node))
        if overwrite | (not 'time' in tree.nodes[current_node]):
            tree.nodes[current_node]['time'] = tree.nodes[parent]['time'] + tree.nodes[current_node]['time_to_parent']

    children = tree.successors(current_node)
    for child in children:
        add_node_times_from_division_times(tree, child)

    return


def remove_times(tree):
    """
    Removes time annotations from nodes and edges of a tree
    """
    for node in tree.nodes:
        tree.nodes[node].pop('time_to_parent')
        if 'time' in tree.nodes[node]:
            tree.nodes[node].pop('time')
    for edge in tree.edges:
        tree.edges[edge].pop('time')
    return


def get_leaves(tree, include_root = True):
    """
    Returns a list of the leaf nodes of a tree
    including the root
    """
    leaves = [node for node in tree if tree.degree(node) <= 1]
    string_leaves = [l for l in leaves if (type(l) == str) and (l != 'root')]
    number_leaves = [l for l in leaves if isinstance(l, Number)]

    number_leaves.sort()
    string_leaves.sort()
    if include_root & ('root' in tree):
        string_leaves = string_leaves + ['root']

    return number_leaves + string_leaves


def get_internal_nodes(tree):
    """
    Returns a list of the non-leaf nodes of a tree
    """
    nodes = [node for node in tree if tree.degree(node) >= 2]
    nodes.sort()
    return nodes

def get_leaf_descendants(tree, node):
    """
    Returns a list of the leaf nodes of the tree that are
    descendants of node
    """
    if tree.out_degree(node) == 0:
        return [node]
    else:
        children = tree.successors(node)
        leaf_descendants = []
        for child in children:
            leaf_descendants = leaf_descendants + get_leaf_descendants(tree, child)
        return leaf_descendants
    return

def compute_leaf_times(tree, num_leaves):
    """
    Computes the list of times of the leaves by adding 'time_to_parent' along the path to 'root'
    """
    rna,  barcodes = extract_data_arrays(tree)
    times = np.zeros(barcodes.shape[0])
    for leaf in range(barcodes.shape[0]):
        current_node = leaf
        while not current_node == 'root':
            times[leaf] = times[leaf] + tree.nodes[current_node]['time_to_parent']
            current_node = next(tree.predecessors(current_node))

    return times

def extract_data_arrays(tree):
    """
    Returns arrays of the RNA expression and barcodes from leaves of the tree

    Each row of each array is a cell
    """

    leaves = get_leaves(tree, include_root = False)
    expressions = np.array([tree.nodes[leaf]['cell'].x for leaf in leaves])
    barcodes = np.array([tree.nodes[leaf]['cell'].barcode for leaf in leaves])
    return expressions, barcodes


def extract_ancestor_data_arrays(late_tree, time, params):
    """
    Returns arrays of the RNA expression and barcodes for ancestors of leaves of the tree
    
    Each row of each array is a leaf node
    """
    
    leaves = get_leaves(late_tree, include_root = False)


    early_tree = truncate_tree(late_tree, time, params)
    cells_early = get_leaves(early_tree, include_root = False)
    
    
    expressions = np.nan*np.ones([len(leaves), params.num_genes])
    barcodes = np.nan*np.ones([len(leaves), params.barcode_length])

    for cell in cells_early:
        parent = next(early_tree.predecessors(cell))
        late_tree_cell = None
        for child in late_tree.successors(parent):
            if late_tree.nodes[child]['cell'].seed == early_tree.nodes[cell]['cell'].seed:
                late_tree_cell = child
                break
        if late_tree_cell == None:
            raise ValueError("A leaf in early_tree does not appear in late_tree. Cannot find coupling." +
                             "\nCheck whether either tree has been modified since truncating.")

        descendants = get_leaf_descendants(late_tree, late_tree_cell)

        expressions[descendants, :] = early_tree.nodes[cell]['cell'].x
        barcodes[descendants, :] = early_tree.nodes[cell]['cell'].barcode

    return expressions, barcodes
    













# Truncating a tree at an earlier endpoint


def truncate_tree(tree, new_end_time, params, inplace = False, current_node = 'root', next_leaf_to_add = 0):
    """
    Removes all nodes at times greater than new_end_time
    and adds new leaves at exactly new_end_time

    params: simulation parameters used to create tree
    """
    if not inplace:
        tree = copy.deepcopy(tree)

    if not ('time' in tree.nodes['root']):
        add_node_times_from_division_times(tree)

    if tree.nodes[current_node]['time'] >= new_end_time:
        parent = next(tree.predecessors(current_node))

        initial_cell = tree.nodes[parent]['cell'].deepcopy()
        initial_cell.seed = tree.nodes[current_node]['cell'].seed

        new_cell = sim.evolve_cell(initial_cell, 
                               new_end_time - tree.nodes[parent]['time'],
                               params)

        remove_node_and_descendants(tree, current_node)
        
        tree.add_node(next_leaf_to_add)
        tree.nodes[next_leaf_to_add]['time'] = new_end_time
        tree.nodes[next_leaf_to_add]['time_to_parent'] = new_end_time - tree.nodes[parent]['time']
        tree.nodes[next_leaf_to_add]['cell'] = new_cell

        tree.add_edge(parent, next_leaf_to_add, time = new_end_time - tree.nodes[parent]['time'])
        next_leaf_to_add = next_leaf_to_add + 1
    else:
        # not just tree.successors(node) because that changes as nodes are removed
        children = [child for child in tree.successors(current_node)]
        for child in children:
            tree, next_leaf_to_add = truncate_tree(tree,
                                                   new_end_time,
                                                   params,
                                                   inplace = True,
                                                   current_node = child,
                                                   next_leaf_to_add = next_leaf_to_add)
    
    if inplace:
        # proxy for "if in recursive case"
        return tree, next_leaf_to_add
    else:
        return tree


def remove_node_and_descendants(tree, node):
    """
    Removes a node and all its descendants from the tree
    """

    # not just tree.successors(node) because that changes as
    # nodes are removed
    children = [child for child in tree.successors(node)]
    for child in children:
        remove_node_and_descendants(tree, child)

    parent = next(tree.predecessors(node))

    tree.remove_edge(parent, node)
    tree.remove_node(node)

    return tree




def resample_cells(tree, params, current_node = 'root', inplace = False):
    """
    Runs a new simulation of the cell evolution on a fixed tree
    """
    if not inplace:
        tree = copy.deepcopy(tree)

    for child in tree.successors(current_node):
        initial_cell = tree.nodes[current_node]['cell'].deepcopy()
        initial_cell.reset_seed()
        
        tree.nodes[child]['cell'] = sim.evolve_cell(initial_cell,
                                               tree.nodes[child]['time_to_parent'],
                                               params)
        resample_cells(tree, params, current_node = child, inplace = True)
        
    return tree


def get_true_coupling(early_tree, late_tree):
    """
    Returns the coupling between leaves of early_tree and their descendants in
    late_tree. Assumes that early_tree is a truncated version of late_tree

    The marginal over the early cells is uniform; if cells have different
    numbers of descendants, the marginal over late cells will not be uniform.
    """
    num_cells_early = len(get_leaves(early_tree)) - 1
    num_cells_late = len(get_leaves(late_tree)) - 1
    
    coupling = np.zeros([num_cells_early, num_cells_late])
    
    cells_early = get_leaves(early_tree, include_root = False)
    
    
    for cell in cells_early:
        parent = next(early_tree.predecessors(cell))
        late_tree_cell = None
        for child in late_tree.successors(parent):
            if late_tree.nodes[child]['cell'].seed == early_tree.nodes[cell]['cell'].seed:
                late_tree_cell = child
                break
        if late_tree_cell == None:
            raise ValueError("A leaf in early_tree does not appear in late_tree. Cannot find coupling." +
                             "\nCheck whether either tree has been modified since truncating.")
        descendants = get_leaf_descendants(late_tree, late_tree_cell)
        coupling[cell, descendants] = 1/(num_cells_early*len(descendants))
    
    return coupling
    
    
    
def get_lineage_distances_across_time(early_tree, late_tree):
    """                                                                                                                                                                                                     
    Returns the matrix of lineage distances between leaves of early_tree and leaves in                                                                                                                              
    late_tree. Assumes that early_tree is a truncated version of late_tree                                                                                                                                  
    """
    num_cells_early = len(get_leaves(early_tree)) - 1
    num_cells_late = len(get_leaves(late_tree)) - 1

    d = np.zeros([num_cells_early, num_cells_late])

    cells_early = get_leaves(early_tree, include_root = False)
    cells_late = get_leaves(late_tree, include_root = False)


    # get length of path up to parent of early_cell and back to early_cell
    for late_cell in cells_late:
        distance_dictionary, tmp = nx.single_source_dijkstra(nx.Graph(late_tree),
                                                            late_cell,
                                                            weight = 'time')
        for early_cell in cells_early:
            d[early_cell, late_cell] = (distance_dictionary[next(early_tree.predecessors(early_cell))]
                                        + early_tree.nodes[early_cell]['time_to_parent'])

    # correct distances to descendants
    for early_cell in cells_early:
        parent = next(early_tree.predecessors(early_cell))
        late_cell = None
        for child in late_tree.successors(parent):
            if late_tree.nodes[child]['cell'].seed == early_tree.nodes[early_cell]['cell'].seed:
                late_cell = child
                break
        if late_cell is not None:
            descendants = get_leaf_descendants(late_tree, late_cell)
            d[early_cell, descendants] = (d[early_cell, descendants]
                                          - 2*early_tree.nodes[early_cell]['time_to_parent'])

    return d

























# For OT using the lineage information, we want to have nodes for ancestors at the early sampling time

def add_nodes_at_time(tree, time_to_add, current_node = 'root', num_nodes_added = 0):
    """
    Splits every edge (u,v) where u['time'] < time_to_add < v['time']

    into (u, w) and (w, v) with w['time'] = time_to_add

    Newly added nodes {w} are labeled as tuples (time_to_add, i)

    The input tree should be annotated with node times already
    """

    if tree.nodes[current_node]['time'] == time_to_add:
        # Do not add a new node if we already have one at the correct time
        return num_nodes_added
    else:
        current_children = [child for child in tree.successors(current_node)]
        # Not iterating over edges directly because we're adding edges along the way
        for child in current_children:
            if tree.nodes[child]['time'] > time_to_add:
                split_edge(tree, (current_node, child), (time_to_add, num_nodes_added))
                
                # Add annotations and correct old ones
                tree.nodes[(time_to_add, num_nodes_added)]['time'] = time_to_add
                tree.nodes[(time_to_add, num_nodes_added)]['time_to_parent'] = time_to_add - tree.nodes[current_node]['time']
                tree.nodes[child]['time_to_parent'] = tree.nodes[child]['time'] - time_to_add
                tree.edges[current_node, (time_to_add, num_nodes_added)]['time'] = time_to_add - tree.nodes[current_node]['time']
                tree.edges[(time_to_add, num_nodes_added), child]['time'] = tree.nodes[child]['time'] - time_to_add
                
                num_nodes_added = num_nodes_added + 1
            elif tree.nodes[child]['time'] <= time_to_add:
                num_nodes_added = add_nodes_at_time(tree,
                                                    time_to_add,
                                                    current_node = child,
                                                    num_nodes_added = num_nodes_added)

        return num_nodes_added



def split_edge(tree, edge, new_node):
    tree.remove_edge(edge[0], edge[1])
    tree.add_node(new_node)
    tree.add_edge(edge[0], new_node)
    tree.add_edge(new_node, edge[1])
    return




def get_components(graph, edge_length_key = 'time'):
    """
    Returns subgraph views corresponding to connected components of the graph
    if edges of infinite length are removed
    
    Parameters
    ----------
    graph: NetworkX graph
    edge_length_key: default 'time'

    Returns
    -------
    subgraphs: List of NetworkX subgraph views
    """
    # copying graph as undirected
    g = nx.Graph(graph)
    
    edges_to_remove = [e for e in g.edges if g.edges[e][edge_length_key] == np.inf]
    g.remove_edges_from(edges_to_remove)
    
    component_node_sets = nx.connected_components(g)

    return [graph.subgraph(component_nodes) for component_nodes in component_node_sets] 












def add_conditional_means_and_variances(tree, observed_nodes):
    """
    Adds the mean and variance of the posterior on 'x' for each of the unobserved
    nodes, conditional on the observed values of 'x' in observed_nodes,
    assuming that differences along edges are Gaussian with variance equal to 
    the length of the edge.

    In doing so, also adds inverse time annotations to edges.

    If no nodes in tree are observed, inverse time annotations are added but
    conditional means and variances are not (as there is nothing to condition on).
    """
    node_list = [n for n in tree.nodes]
    
    add_inverse_times_to_edges(tree)
    l = nx.laplacian_matrix(nx.Graph(tree), nodelist = node_list, weight = 'inverse time')
    
    unobserved_nodes = [n for n in node_list if not n in observed_nodes]
    # Resorting so the order of indices in all matrices match
    # and removing observed_nodes not in tree
    observed_nodes = [n for n in node_list if n in observed_nodes]
    if len(observed_nodes) == 0:
        # no observed nodes in this component
        return
    unobserved_node_indices = [i for i in range(len(node_list)) if not node_list[i] in observed_nodes]
    observed_node_indices = [i for i in range(len(node_list)) if node_list[i] in observed_nodes]
    
    conditioned_precision = np.array(l[np.ix_(unobserved_node_indices, unobserved_node_indices)].todense())
    conditioned_covariance = np.linalg.inv(conditioned_precision)
    
    x_differences = np.array([tree.nodes[n]['cell'].x - tree.nodes[observed_nodes[0]]['cell'].x
                              for n in observed_nodes])
    
    conditioned_means = -1*(conditioned_covariance@l[np.ix_(unobserved_node_indices, observed_node_indices)]@x_differences)
    
    for n in observed_nodes:
        tree.nodes[n]['x mean'] = tree.nodes[n]['cell'].x
        tree.nodes[n]['x variance'] = 0
        
    for i, n in zip(range(len(unobserved_nodes)), unobserved_nodes):
        tree.nodes[n]['x mean'] = conditioned_means[i, :] + tree.nodes[observed_nodes[0]]['cell'].x
        tree.nodes[n]['x variance'] = conditioned_covariance[i, i] # assumed isotropic
    
    return
        


def get_ancestor_data(tree, time, leaf = None):
    if leaf == None:
        # get all leaf ancestors
        leaves = [l for l in get_leaves(tree) if not l == 'root']
        data = [get_ancestor_data(tree, time, l) for l in leaves]
        return np.array([d[0] for d in data]), np.array([d[1] for d in data])
    else:
        current_node = leaf
        while tree.nodes[current_node]['time'] > time:
            current_node = next(tree.predecessors(current_node))
        if tree.nodes[current_node]['time'] < time:
            error = "Tree has no ancestor of cell " + str(leaf) + " at time " + str(time)
            raise ValueError(error)
        return (tree.nodes[current_node]['x mean'], tree.nodes[current_node]['x variance'])
    return









##########################
# Tree fitting functions #
##########################

def make_tree_from_nonnested_clones(clone_matrix, time, root_time_factor = 1000):
    """
    Creates a forest of stars from clonally-labeled data. The centers of the stars are connected to a root far in the past.

    Parameters
    ----------
    clone_matrix: Boolean array with shape [num_cells, num_clones]
        Each entry is 1 if the corresponding cell belongs to the corresponding clone and zero otherwise.
        Each cell should belong to exactly one clone.
    time: Number
        The time of sampling of cells relative to initial clonal labelling.
    root_time_factor: Number, default 1000
        Relative time to root of tree (i.e., most recent common ancestor of all cells).
        The time of the root is set to -root_time_factor*time.
        The default is large so minimal information is shared across clones.
    Returns
    -------
    fitted_tree: NetworkX DiGraph
        A tree annotated with edge and node times
    """
    fitted_tree = nx.DiGraph()

    n_cells, n_clones = clone_matrix.shape
    clone_labels = [-i-1 for i in range(n_clones) if np.sum(clone_matrix[:, i]) > 0] # negative so distinct from cells, skipping empty clones
    
    # add nodes for cells
    fitted_tree.add_nodes_from(range(n_cells), time = time, time_to_parent = time)

    # add nodes for clone progenitors
    fitted_tree.add_nodes_from(clone_labels, time = 0, time_to_parent = root_time_factor*time)

    # add edges from clone progenitors to observed cells
    for clone in clone_labels:
        cells_in_clone = np.nonzero(clone_matrix[:, -clone - 1])[0]
        fitted_tree.add_edges_from([(clone, cell, {'time': time}) for cell in cells_in_clone])

    # add root
    fitted_tree.add_node('root', time = -root_time_factor*time)

    # add edges to root
    fitted_tree.add_edges_from([('root', i, {'time': root_time_factor*time}) for i in clone_labels])

    return fitted_tree


def make_tree_from_clones(clone_matrix, time, clone_times, root_time = -np.inf):
    """
    Adds a leaf for each row in clone_matrix to clone_reference_tree. The parent is set as the
    clone that the cell is a member of with the latest labeling time.

    clone_reference_tree is edited in place rather than returned.

    Parameters
    ----------
    clone_matrix: Boolean array with shape [num_cells, num_clones]
        Each entry is 1 if the corresponding cell belongs to the corresponding clone and zero otherwise.
    clone_times: Vector of length num_clones
        Each entry has the time of labeling of the corresponding clone.
    time: Number
        The time of sampling of cells.
    root_time: Number, default -np.inf
        The time of the most recent common ancestor of all clones. 
        If -np.inf, clones are effectively treated as unrelated

    Returns
    -------
    fitted_tree: NetworkX DiGraph
        A tree annotated with edge and node times
    """
    fitted_tree = make_clone_reference_tree(clone_matrix, clone_times, root_time)
    add_samples_to_clone_tree(clone_matrix, clone_times, fitted_tree, time)
    return fitted_tree


def add_samples_to_clone_tree(clone_matrix, clone_times, clone_reference_tree, sampling_time):
    """
    Adds a leaf for each row in clone_matrix to clone_reference_tree. The parent is set as the
    clone that the cell is a member of with the latest labeling time.

    clone_reference_tree is edited in place rather than returned.

    Parameters
    ----------
    clone_matrix: Boolean array with shape [num_cells, num_clones]
        Each entry is 1 if the corresponding cell belongs to the corresponding clone and zero otherwise.
    clone_times: Vector of length num_clones
        Each entry has the time of labeling of the corresponding clone.
    clone_reference_tree: 
        The tree of lineage relationships among clones.
    sampling_time: Number
        The time of sampling of the cells. Should be greater than all clone labeling times.

    Returns
    -------
    """
    n_cells, n_clones = clone_matrix.shape
    for cell in range(n_cells):
        parent_index = get_parent_clone_of_leaf(cell, clone_matrix, clone_times)
        parent_label = 'clone_' + str(parent_index)
        edge_time = sampling_time - clone_reference_tree.nodes[parent_label]['time']
        clone_reference_tree.add_node(cell, time = sampling_time, time_to_parent = edge_time)
        clone_reference_tree.add_edge(parent_label, cell, time = edge_time)
    return


def get_parent_clone_of_leaf(leaf, clone_matrix, clone_times):
    """
    Returns the index of the clone that the leaf is a member of with the latest labeling time.
    """
    candidate_clones = np.nonzero(clone_matrix[leaf, :])[0]
    candidate_times = clone_times[candidate_clones]
    maximal_time_indices = np.nonzero(candidate_times == np.max(candidate_times))[0]

    if len(maximal_time_indices) != 1:
        raise ValueError("Could not unambiguously assign a sampled cell to a parent clone.")
    else:
        return candidate_clones[maximal_time_indices[0]]


def make_clone_reference_tree(clone_matrix, clone_times, root_time = -np.inf):
    """
    Makes a tree with nodes for each clone.
    
    Parameters
    ----------
    clone_matrix: Boolean array with shape [num_cells, num_clones]
        Each entry is 1 if the corresponding cell belongs to the corresponding clone and zero otherwise.
    clone_times: Vector of length num_clones
        Each entry has the time of labeling of the corresponding clone.
    root_time: Number, default -np.inf
        The time of the most recent common ancestor of all clones. 
        If -np.inf, clone subtrees are effectively treated independently.

    Returns
    -------
    clone_reference_tree: NetworkX DiGraph
        A tree of clones (not sampled cells), annotated with edge and node times
    """

    n_cells, n_clones = clone_matrix.shape
    unique_times = np.unique(clone_times) # output is sorted
    clone_labels = ['clone_' + str(i) for i in range(n_clones)]

    # make base level of tree
    clone_reference_tree = nx.DiGraph()
    clone_reference_tree.add_node('root', time = root_time)

    current_clone_indices = np.nonzero(clone_times == unique_times[0])[0]
    current_clone_labels = [clone_labels[i] for i in current_clone_indices]
    clone_reference_tree.add_nodes_from(current_clone_labels, time = unique_times[0], time_to_parent = unique_times[0] - root_time)
    clone_reference_tree.add_edges_from([('root', label, {'time': unique_times[0] - root_time}) for label in current_clone_labels])

    # make subsequent levels
    for level, labeling_time in zip(range(1, len(unique_times)), unique_times[1:]):
        current_clone_indices = np.nonzero(clone_times == labeling_time)[0]
        current_clone_labels = [clone_labels[i] for i in current_clone_indices]

        clone_reference_tree.add_nodes_from(current_clone_labels, time = labeling_time)

        for clone_index in current_clone_indices:
            parent_index = find_parent_clone(clone_index, clone_matrix, clone_times)
            if parent_index == 'root':
                parent_label = 'root'
            else:
                parent_label = clone_labels[parent_index]

            time_to_parent = labeling_time - clone_reference_tree.nodes[parent_label]['time']
            
            clone_reference_tree.nodes[clone_labels[clone_index]]['time_to_parent'] = time_to_parent
            clone_reference_tree.add_edge(parent_label, clone_labels[clone_index], time = time_to_parent)

    return clone_reference_tree


def find_parent_clone(clone, clone_matrix, clone_times):
    """
    Returns the parent of a subclone, assuming this is uniquely defined as the clone from
    an earlier time point whose barcode was observed in a cell from the subclone.

    Parameters
    ----------
    clone: int
        Index of clone whose parent will be returned
    clone_matrix: Boolean array with shape [num_cells, num_clones]
        Each entry is 1 if the corresponding cell belongs to the corresponding clone and zero otherwise.
    clone_times: Vector of length num_clones
        Each entry has the time of labeling of the corresponding clone.

    Returns
    -------
    parent: int
       Index of parent clone.
    """
    clone_is_earlier = clone_times < clone_times[clone]
    clone_is_shared = np.array(np.sum(clone_matrix[np.nonzero(clone_matrix[:, clone] == 1)[0], :], 0) > 0)
    clone_is_shared = np.reshape(clone_is_shared, clone_is_shared.size) # does nothing if everyhing was ndarray, but corrects an issue with sparse matrices

    candidates = np.nonzero(clone_is_earlier & clone_is_shared)[0]

    if len(candidates) == 1:
        return candidates[0]
    elif len(candidates) == 0:
        # no parent clone observed
        return 'root'
    else:
        print(clone_is_earlier)
        print(clone_is_shared)
        print(clone, candidates)
        raise ValueError("The input clone matrix does not unambiguously define a clonal lineage tree. Please use another method to create the tree.")


###################################
# Neighbor joining implementation #
###################################

class NeighborJoinNode:
    def __init__(self, subtree, subtree_root, has_global_root):
        self.subtree = subtree
        self.subtree_root = subtree_root
        self.has_global_root = has_global_root
        return




def neighbor_join(distance_matrix):
    """
    Creates a tree by neighbor joining with the input distance matrix

    Final row/column of distance_matrix assumed to correspond to the root
    (unmutated) barcode
    """
    n = distance_matrix.shape[0]

    assert(n >= 3) # no need to bother with neighbor joining if one or two nodes

    unjoined_nodes = []
    for i in range(n-1):
        subtree = nx.DiGraph()
        subtree.add_node(i)
        node = NeighborJoinNode(subtree, i, False)
        unjoined_nodes.append(node)

    subtree = nx.DiGraph()
    subtree.add_node('root')
    node = NeighborJoinNode(subtree, 'root', True)
    unjoined_nodes.append(node)

    next_node_to_add = -1 # integer label of next node to add, will count down
    # (leaf nodes have positive integer labels, root is labeled 'root')

    while len(unjoined_nodes) > 3:
        # recursive part of neighbor joining algorithm

        # Compute Q matrix
        Q = compute_q_matrix(distance_matrix)

        # Pick nodes to join
        nodes_to_join = pick_joined_nodes(Q)
        
        distances = distances_to_joined_node(distance_matrix, nodes_to_join)

        # Remove those nodes and add merged node
        new_node = join_nodes(unjoined_nodes[nodes_to_join[0]], unjoined_nodes[nodes_to_join[1]], next_node_to_add, distances)
        next_node_to_add = next_node_to_add - 1
        
        unjoined_nodes = [unjoined_nodes[i] for i in range(len(unjoined_nodes)) if i not in nodes_to_join]
        unjoined_nodes.append(new_node)

        # Compute new distances
        distance_matrix = compute_new_distances(distance_matrix, nodes_to_join)



    # only three remaining nodes now; join an arbitrary two of them directly
    distances = distances_to_joined_node(distance_matrix, [0,1])
    last_join = join_nodes(unjoined_nodes[0], unjoined_nodes[1], next_node_to_add, distances)
    next_node_to_add = next_node_to_add - 1
    
    last_edge_distance = (distance_matrix[2, 0] 
                          + distance_matrix[2, 1]
                          - distance_matrix[1, 0])/2
    fitted_tree = nx.compose(unjoined_nodes[2].subtree, last_join.subtree)
    
    # add 'time_to_parent' to the node that is not the root
    if last_join.has_global_root:
        assert(not unjoined_nodes[2].has_global_root)
        fitted_tree.add_edge(last_join.subtree_root, unjoined_nodes[2].subtree_root)
        fitted_tree.nodes[unjoined_nodes[2].subtree_root]['time_to_parent'] = last_edge_distance
    else:
        assert(unjoined_nodes[2].has_global_root)
        fitted_tree.add_edge(unjoined_nodes[2].subtree_root, last_join.subtree_root)
        fitted_tree.nodes[last_join.subtree_root]['time_to_parent'] = last_edge_distance

    
    return fitted_tree


        

def distances_to_joined_node(distance_matrix, nodes_to_join):
    pair_distance = distance_matrix[nodes_to_join[0], nodes_to_join[1]]
    n = distance_matrix.shape[0]

    d1 = pair_distance/2 + (np.sum(distance_matrix[:, nodes_to_join[0]])
                            - np.sum(distance_matrix[:, nodes_to_join[1]])
                            )/ (2*(n-2))

    d2 = pair_distance - d1

    return d1, d2



def join_nodes(node1, node2, new_root, distances):
    T = nx.compose(node1.subtree, node2.subtree)
    T.add_node(new_root)

    if node1.has_global_root:
        parent = node1.subtree_root
        child = new_root
    else:
        parent = new_root
        child = node1.subtree_root
    T.add_edge(parent, child)
    T.nodes[child]['time_to_parent'] = distances[0]

    if node2.has_global_root:
        parent = node2.subtree_root
        child = new_root
    else:
        parent = new_root
        child = node2.subtree_root

    T.add_edge(parent, child)
    T.nodes[child]['time_to_parent'] = distances[1]

    return NeighborJoinNode( T, new_root,  node1.has_global_root | node2.has_global_root)



def compute_q_matrix(distance_matrix):
    """
    Computes the Q-matrix for neighbor joining
    """

    n = distance_matrix.shape[0]

    Q = (n-2)*distance_matrix
    Q = Q - np.sum(distance_matrix, axis = 0)
    Q = (Q.transpose() - np.sum(distance_matrix, 1)).transpose()
    # never pick to merge a node with itself
    Q[range(n), range(n)] = np.inf
    return Q


def pick_joined_nodes(Q):
    """
    In default neighbor joining, returns the indices of the pair
    of nodes with the lowest Q value

    TODO: extend to allow stochastic neighbor joining
    """
    return np.unravel_index(np.argmin(Q), Q.shape)


def compute_new_distances(distance_matrix, nodes_to_join):
    
    n = distance_matrix.shape[0]
    kept_nodes = [i for i in range(n) if i not in nodes_to_join]

    # Leave distances between unmerged nodes unchanged
    new_distances = np.zeros([n-1, n-1])
    new_distances[:(n-2), :(n-2)] = distance_matrix[np.ix_(kept_nodes, kept_nodes)]

    distances_to_new_node = (distance_matrix[nodes_to_join[0], :]
                             + distance_matrix[nodes_to_join[1], :]
                             - distance_matrix[nodes_to_join[0],
                                               nodes_to_join[1]]
                             )/2
    distances_to_new_node = distances_to_new_node[kept_nodes]

    new_distances[n-2, 0:(n-2)] = distances_to_new_node
    new_distances[0:(n-2), n-2] = distances_to_new_node

    return new_distances
    





def tree_accuracy(tree1, tree2):
    """
    Returns the fraction of nontrivial splits appearing
    in both trees
    """
    return 1 - tree_discrepancy(tree1, tree2)

def tree_discrepancy(tree1, tree2):
    """
    Computes a version of the Robinson-Foulds distance
    between two trees rescaled to be between 0 and 1
    """
    n_leaves = len(get_leaves(tree1))
    d = robinson_foulds(tree1, tree2)
    return d/(2*(n_leaves - 3))


def robinson_foulds(tree1, tree2):
    """
    Computes the Robinson-Foulds distance between two trees
    """
    leaves1 = get_leaves(tree1)
    leaves2 = get_leaves(tree2)

    # Robinson-Foulds is only defined if the trees have the same leaf set
    assert(leaves1 == leaves2)

    ete3_t1 = tree_to_ete3(tree1)
    ete3_t2 = tree_to_ete3(tree2)
    return ete3_t1.robinson_foulds(ete3_t2)[0]








####################################
# Converting trees between formats #
####################################

def convert_newick_to_networkx(newick_tree, leaf_labels, leaf_time = None, root_label = 'root', unlabeled_nodes_added = 0, at_global_root = True):
    """
    Converts a tree from the Newick package's format to LineageOT's annotated NetworkX DiGraph. Ignores existing annotations, except edge lengths.

    Parameters
    ----------
    newick_tree : newick.Node or [newick.Node]
        A tree loaded by the Newick package.
    leaf_labels : list
        The label of each leaf in the Newick tree, sorted to align with the gene expression AnnData object filtered to cells at the corresponding time
    leaf_time : float (default None)
        The time of sampling of the leaves. If unspecified, the root of the tree is assigned time 0.
    root_label : str (default 'root')
        The label of the root node of the tree
    unlabeled_nodes_added : int (default 0)
        The number of previously-unlabeled nodes that have already been added to the tree. Leave as 0 for any top-level use of the function.
    at_global_root : bool (default True)
        Whether the function is being called to convert a full tree or a subtree.
    Returns
    -------
    tree : Networkx DiGraph
        The saved tree, in LineageOT's format.
        Each node is annotated with 'time_to_parent' and 'time' (which indicates either the time of sampling (for observed cells) or the time of division (for unobserved ancestors)).
        Edges are directed from parent to child and are annotated with 'time' equal to the child node's 'time_to_parent'.
        Observed node indices correspond to their index in leaf_labels.
    """
    assert(type(newick_tree) == newick.Node)
    tree = nx.DiGraph()
    if root_label == None:
        # node was unlabeled in original Newick tree
        # give it a new unique name
        root_label = "Unlabeled_" + str(unlabeled_nodes_added)
        unlabeled_nodes_added = unlabeled_nodes_added + 1
    root_name = root_label
    if len(newick_tree.descendants) == 0:
        # at a leaf, use index of label as name
        if not (root_label in leaf_labels):
            raise ValueError("One of the leaves in the Newick tree does not have a corresponding label in leaf_labels")
        root_label = leaf_labels.index(root_label)
        
    tree.add_node(root_label, name = root_name)
    
    for child in newick_tree.descendants:
        # convert tree for child
        subtree, unlabeled_nodes_added, new_child_label = convert_newick_to_networkx(child, leaf_labels, leaf_time = leaf_time, root_label = child.name, unlabeled_nodes_added = unlabeled_nodes_added, at_global_root = False)
        
        # add to existing tree
        tree = nx.union(tree, subtree) # not sure this is efficient
        
        # annotate edge times
        tree.add_edge(root_label, new_child_label, time=child.length)
        tree.nodes[new_child_label]['time_to_parent'] = child.length

    # annotate node times

    if leaf_time == None:
        # reference time is global root, do nothing in recursive case
        if at_global_root:
            tree.nodes[root_label]['time'] = 0
            add_node_times_from_division_times(tree, current_node = root_label) 
    else:
        if len(newick_tree.descendants) == 0:
            # at leaf
            tree.nodes[root_label]['time'] = leaf_time
        else:
            # add time annotations backwards
            times_from_children = np.array([tree.nodes[child]['time'] - tree.nodes[child]['time_to_parent'] for child in tree.successors(root_label)])
            if np.all(times_from_children == times_from_children[0]):
                # time labels from different branches are consistent
                tree.nodes[root_label]['time'] = times_from_children[0]
            else:
                raise ValueError("Edge length annotations in Newick tree are not consistent with a single sampling time for all leaves. Consider using the global root as a reference time.")

    if at_global_root:
        return tree
    else:
        return tree, unlabeled_nodes_added, root_label


def tree_to_ete3(tree):
    """
    Converts a tree to ete3 format.
    Useful for calculating Robinson-Foulds distance.
    """
    if type(tree) == ete3.coretype.tree.TreeNode:
        return tree
    elif type(tree) == nx.classes.digraph.DiGraph:
        if 'root' in tree.nodes():
            ete3_tree = subtree_to_ete3(tree, 'root')
        else:
            # pick smallest-label leaf as root
            leaves = get_leaves(tree)
            ete3_tree = subtree_to_ete3(tree, leaves[0])
    else:
        raise(NotImplementedError, 'tree_to_ete3 only converts from networkx DiGraph')

    return ete3_tree


def subtree_to_ete3(tree, current_root):
    """
    Converts the subtree from current_root to ete3 format
    """
    assert(type(tree) == nx.classes.digraph.DiGraph)
    
    ete3_tree = ete3.Tree()
    ete3_tree.name = current_root

    for subtree_root in tree.successors(current_root):
        ete3_subtree = subtree_to_ete3(tree, subtree_root)
        if 'time_to_parent' in tree.nodes[subtree_root]:
            d = tree.nodes[subtree_root]['time_to_parent']
        else:
            d = None

        ete3_tree.add_child(ete3_subtree, dist = d)

    return ete3_tree




def cvxopt_qp_from_numpy(P, q, G, h):
    """
    Converts arguments to cvxopt matrices and runs
    cvxopt's quadratic programming solver
    """

    P = cvxopt_matrix(P)
    q = cvxopt_matrix(q)
    G = cvxopt_matrix(G)
    h = cvxopt_matrix(np.squeeze(h))

    sol = cvxopt_qp(P, q, G, h)
    return np.array(sol['x'])




def compute_tree_distances(tree):
    """
    Computes the matrix of pairwise distances between leaves of the tree
    """
    num_leaves = len(get_leaves(tree)) - 1
    distances = np.zeros([num_leaves, num_leaves])
    for leaf in range(num_leaves):
        distance_dictionary, tmp = nx.multi_source_dijkstra(tree.to_undirected(), [leaf], weight = 'time')
        for target_leaf in range(num_leaves):
            distances[leaf, target_leaf] = distance_dictionary[target_leaf]
    return distances




import anndata
import newick
import numpy as np
import os
import ot

import lineageot.inference as inf


def fit_tree(adata, time, barcodes_key = 'barcodes', clones_key = "X_clone", clone_times = None, method = 'neighbor join'):
    """
    Fits a lineage tree to lineage barcodes of all cells in adata. To compute the lineage tree for a specific time point,
    filter adata before calling fit_tree. The fitted tree is annotated with node times but not states.

    Parameters
    ----------
    adata : AnnData
        Annotated data matrix with lineage-traced cells
    time : Number
        Time of sampling of the cells of adata relative to most recent common ancestor (for dynamic lineage tracing) or labeling time (for static lineage tracing).
    barcodes_key : str, default 'barcodes'
        Key in adata.obsm containing cell barcodes. Ignored if using clonal data.
        If using barcode data, each row of adata.obsm[barcodes_key] should be a barcode where each entry corresponds to a possibly-mutated site.
        A positive number indicates an observed mutation, zero indicates no mutation, and -1 indicates the site was not observed.
    clones_key: str, default 'X_clone'
        Key in adata.obsm containing clonal data. Ignored if using barcodes directly.
        If using clonal data, adata.obsm[clones_key] should be a num_cells x num_clones boolean matrix.
        Each entry is 1 if the corresponding cell belongs to the corresponding clone and zero otherwise.
    clone_times: Vector of length num_clones, default None
        Ignored unless method is 'clones'.
        Each entry contains the time of labeling of the corresponding column of adata.obsm[clones_key].
    method : str
        Inference method used to fit tree.
        Current options are 'neighbor join' (for barcodes from dynamic lineage tracing), 'non-nested clones' (for non-nested clones from static lineage tracing), or 'clones' (for possibly-nested clones from static lineage tracing).

    Returns
    -------
    tree : Networkx DiGraph
        A fitted lineage tree. 
        Each node is annotated with 'time_to_parent' and 'time' (which indicates either the time of sampling (for observed cells) or the time of division (for unobserved ancestors)). 
        Edges are directed from parent to child and are annotated with 'time' equal to the child node's 'time_to_parent'.
        Observed node indices correspond to their row in adata.
    """

    if method == "neighbor join":
        # compute distances
        barcode_length = adata.obsm[barcodes_key].shape[1]
        # last row is (unobserved) root of the tree
        lineage_distances = inf.barcode_distances(np.concatenate([adata.obsm[barcodes_key], np.zeros([1,barcode_length])]))
        
        # compute tree
        fitted_tree = inf.neighbor_join(lineage_distances)
        
        # annotate tree with node times
        inf.add_leaf_barcodes(fitted_tree, adata.obsm[barcodes_key])
        inf.add_leaf_times(fitted_tree, time)
        
        # Estimating a uniform mutation rate for all target sites
        rate_estimate = inf.rate_estimator(adata.obsm[barcodes_key], time)
        inf.annotate_tree(fitted_tree, 
                          rate_estimate*np.ones(barcode_length),
                          time_inference_method = 'least_squares');
    elif method == "non-nested clones":
        # check to confirm clones are not nested
        if not np.all(np.sum(adata.obsm[clones_key], 1) == 1):
            raise ValueError("The tree fitting method 'non-nested clones' assumes each cell is a member of exactly one clone. This is not the case for your data.")

        fitted_tree = inf.make_tree_from_nonnested_clones(adata.obsm[clones_key], time)

    elif method == "clones":
        if clone_times is None:
            raise ValueError("clone_times must be specified in order to fit a tree to nested clones.")
        clone_times = np.array(clone_times) # allowing clone_times to be passed as a raw list without causing errors later
        fitted_tree = inf.make_tree_from_clones(adata.obsm[clones_key], time, clone_times)
    else:
        raise ValueError("'" + method + "' is not an available method for fitting trees.")

    return fitted_tree


def read_newick(filename, leaf_labels, leaf_time = None):
    """
    Loads a tree saved in Newick format and adds annotations required for LineageOT.
    
    Parameters
    ----------
    filename : str
        The name of the file to load from.
    leaf_labels : list
        The label of each leaf in the Newick tree, sorted to align with the gene expression AnnData object filtered to cells at the corresponding time.
    leaf_time : float (default None)
        The time of sampling of the leaves. If unspecified, the root of the tree is assigned time 0.
    Returns
    -------
    tree : Networkx DiGraph
        The saved tree, in LineageOT's format.
        Each node is annotated with 'time_to_parent' and 'time' (which indicates either the time of sampling (for observed cells) or the time of division (for unobserved ancestors)). 
        Edges are directed from parent to child and are annotated with 'time' equal to the child node's 'time_to_parent'.
        Observed node indices correspond to their index in leaf_labels, which should match their row in the gene expression AnnData object filtered to cells at the corresponding time.
    """
    newick_tree = newick.read(filename)[0]
    tree = inf.convert_newick_to_networkx(newick_tree, leaf_labels, leaf_time)
    return tree



def fit_lineage_coupling(adata, time_1, time_2, lineage_tree_t2, time_key = 'time', state_key = None, epsilon = 0.05, normalize_cost = True, ot_method = 'sinkhorn', marginal_1 = [], marginal_2 = [], balance_reg = np.inf):
    """
    Fits a LineageOT coupling between the cells in adata at time_1 and time_2. 
    In the process, annotates the lineage tree with observed and estimated cell states.

    Parameters
    ----------
    adata : AnnData
        Annotated data matrix
    time_1 : Number
        The earlier time point in adata. All times are relative to the root of the tree.
    time_2 : Number
        The later time point in adata. All times are relative to the root of the tree.
    lineage_tree_t2 : Networkx DiGraph
        The lineage tree fitted to cells at time_2. Nodes should already be annotated with times. Annotations related to cell state will be added. 
    time_key : str (default 'time')
        Key in adata.obs and lineage_tree_t2 containing cells' time labels
    state_key : str (default None)
        Key in adata.obsm containing cell states. If None, uses adata.X.
    epsilon : float (default 0.05)
        Entropic regularization parameter for optimal transport
    normalize_cost : bool (default True)
        Whether to rescale the cost matrix by its median before fitting a coupling. 
        Normalizing this way allows us to choose a reasonable default epsilon for data of any scale
    ot_method : str (default 'sinkhorn')
        Method used for the optimal transport solver. 
        Choose from 'sinkhorn', 'greenkhorn', 'sinkhorn_stabilized' and 'sinkhorn_epsilon_scaling' for balanced transport
        and 'sinkhorn', 'sinkhorn_stabilized', and 'sinkhorn_reg_scaling' for unbalanced transport.
        'sinkhorn' is recommended unless you encounter numerical problems.
        See PythonOT docs for more details.
    marginal_1 : Vector (default [])
        Marginal distribution (relative growth rates) for cells at time 1. If empty, assumed uniform.
    marginal_2 : Vector (default [])
        Marginal distribution (relative growth rates) for cells at time 2. If empty, assumed uniform.
    balance_reg : Number
        Regularization parameter for unbalanced transport. Smaller values allow more flexibility in growth rates. If infinite, marginals are treated as hard constraints.
    
    Returns
    -------
    coupling : AnnData
        AnnData containing the lineage coupling. 
        Cells from time_1 are in coupling.obs, cells from time_2 are in coupling.var, and the coupling matrix is coupling.X
    """

    state_arrays = {}
    if state_key == None:
        state_arrays['early'] = adata[adata.obs[time_key] == time_1].X
        state_arrays['late'] = adata[adata.obs[time_key] == time_2].X
    else:
        state_arrays['early'] = adata[adata.obs[time_key] == time_1].obsm[state_key]
        state_arrays['late'] = adata[adata.obs[time_key] == time_2].obsm[state_key]

    # annotate tree
    inf.add_leaf_x(lineage_tree_t2, state_arrays['late'])


    # Add inferred ancestor nodes and states
    inf.add_nodes_at_time(lineage_tree_t2, time_1)

    observed_nodes = [n for n in inf.get_leaves(lineage_tree_t2, include_root = False)]

    # Split tree into components that share information
    components = inf.get_components(lineage_tree_t2)
    # Add annotations for each component separately
    for comp in components:
        inf.add_conditional_means_and_variances(comp, observed_nodes)

    # collect predicted ancestral states
    ancestor_info = inf.get_ancestor_data(lineage_tree_t2, time_1)

    # change backend of ancestor_info to AnnData ArrayView
    # to match state_arrays['early'], because POT v0.8 requires
    # matching backends
    ancestor_states = anndata.AnnData(X = ancestor_info[0])[:,:].X

    # compute cost matrix
    # (converted to numpy array to match default marginal backend)
    lineageOT_cost = np.array(ot.utils.dist(state_arrays['early'], ancestor_states)@np.diag(ancestor_info[1]**(-1)))

    if normalize_cost:
        lineageOT_cost = lineageOT_cost/np.median(lineageOT_cost)

    # fit coupling
    if balance_reg == np.inf:
        coupling_matrix = ot.sinkhorn(marginal_1, marginal_2, lineageOT_cost, epsilon, method = ot_method)
    else:
        coupling_matrix = ot.unbalanced.sinkhorn_unbalanced(marginal_1, marginal_2, lineageOT_cost, epsilon, balance_reg, method = ot_method)


    # reformat coupling as anndata
    coupling = anndata.AnnData(X = coupling_matrix,
                               obs = adata[adata.obs[time_key] == time_1].obs,
                               var = adata[adata.obs[time_key] == time_2].obs
                               )

    return coupling





def save_coupling_as_tmap(coupling, time_1, time_2, tmap_out):
    """
    Saves a LineageOT coupling for downstream analysis with Waddington-OT. 
    A sequence of saved couplings can be loaded in ``wot`` with 
    ``wot.tmap.TransportMapModel.from_directory(tmap_out)``

    Parameters
    ----------
    coupling : AnnData
        The coupling to save.
    time_1 : Number
        The earlier time point in adata. All times are relative to the root of the tree.
    time_2 : Number
        The later time point in adata. All times are relative to the root of the tree.
    tmap_out : str
        The path and prefix to the save file name.
    """
    # Normalize columns to sum to 1
    col_sums = np.sum(coupling.X, axis = 1)
    coupling.X = coupling.X/col_sums[:, np.newaxis]

    # Add constant relative growth rates for initial cells
    coupling.obs['g0'] = 1
    coupling.obs['g1'] = 1

    # Save
    file_name = tmap_out + '_' + str(time_1) + '_' + str(time_2) + '.h5ad'
    file_dir = os.path.dirname(file_name)
    
    if not os.path.exists(file_dir):
        os.mkdir(file_dir)
    coupling.write(file_name)

    # Change normalization back to what it was.
    coupling.X = coupling.X*col_sums[:, np.newaxis]
    return



# Main file for creating simulated data

import numpy as np
import warnings
import copy


class SimulationParameters:
    """
    Storing the parameters for simulated data
    """
    
    def __init__(self,
                 timestep = 0.01,                              # Time step for Euler integration of SDE
                 diffusion_constant = 0.001,                   # Diffusion constant
                 mean_division_time = 10,                      # Mean time before cell division
                 division_time_distribution = "normal",        # Distribution of cell division times
                 division_time_std = 0,                        # Standard deviation of division times
                 target_num_cells = np.inf,                    # Upper bound on number of observed cells
                 mutation_rate = 1,                            # Rate at which barcodes mutate
                 flow_type = 'bifurcation',                    # Type of flow field simulated
                 x0_speed = 1,                                 # Speed at which cells go through transition
                 barcode_length = 15,                          # Number of elements of the barcode
                 back_mutations = False,                       # Whether barcode elements can mutate multiple times
                 num_genes = 3,                                # Number of genes defining cell state
                 initial_distribution_std = 0,                 # Standard deviation of initial cell distribution
                 alphabet_size = 200,                          # Number of possible mutations for a single barcode element
                 relative_mutation_likelihoods = np.ones(200), # Relative likelihood of each mutation
                 keep_tree = True,                             # Whether simulation output includes tree structure
                 enforce_barcode_reproducibility = True,       # Whether to use reproducible_poisson sampling
                 keep_cell_seeds = True                        # Whether to store seeds to reproduce cell trajectories individually
                 ):

        self.timestep = timestep
        self.diffusion_constant = diffusion_constant
        if flow_type in {'mismatched_clusters', 'convergent', 'partial_convergent'}:
            self.diffusion_matrix = np.diag(np.sqrt(diffusion_constant)*np.ones(num_genes))
            self.diffusion_matrix[0, 0] = 0.005*np.sqrt(x0_speed) #small diffusion in 'time' dimension
        else:
            self.diffusion_matrix = np.sqrt(diffusion_constant)

        self.mean_division_time = mean_division_time
        self.division_time_distribution = division_time_distribution
        self.division_time_std = division_time_std
        self.target_num_cells = target_num_cells

        self.mutation_rate = mutation_rate
        self.flow_type = flow_type
        self.x0_speed = x0_speed
        
        self.barcode_length = barcode_length
        self.back_mutations = back_mutations
        self.num_genes = num_genes
        self.initial_distribution_std = initial_distribution_std
        self.alphabet_size = alphabet_size
        self.keep_tree = keep_tree
        self.enforce_barcode_reproducibility = enforce_barcode_reproducibility
        self.keep_cell_seeds = keep_cell_seeds

        if len(relative_mutation_likelihoods) > alphabet_size:
            warnings.warn('relative_mutation_likelihoods too long: ignoring extra entries')
        elif len(relative_mutation_likelihoods) < alphabet_size:
            raise ValueError('relative_mutation_likelihods must be as long as alphabet_size')
        
        self.relative_mutation_likelihoods = relative_mutation_likelihoods[0:alphabet_size]
        if not self.back_mutations:
            self.relative_mutation_likelihoods[0] = 0

        self.mutation_likelihoods = relative_mutation_likelihoods/sum(relative_mutation_likelihoods)


class Cell:
    """
    Wrapper for (rna expression, barcode) arrays
    """
    def __init__(self, x, barcode, seed = None):
        self.x = x
        self.barcode = barcode

        # Note: to recover a cell's trajectory you
        # need both the seed and initial condition
        #
        # This only stores the seed
        if seed != None:
            self.seed = seed
        else:
            self.seed = np.random.randint(2**32)
        return

    def reset_seed(self):
        self.seed = np.random.randint(2**32)

    def deepcopy(self):
        return copy.deepcopy(self)

    def __repr__(self):
        return '<Cell(%s, %s)>' % (self.x, self.barcode)
    def __str__(self):
        return 'Cell with x: %s\tBarcode: %s' % ( self.x, self.barcode)






def single_bifurcation_flow(x):
    v = np.zeros(x.shape)
    v[0] = 1
    v[1] = -x[1]**3 + x[0]*x[1]
    return v


def mismatched_clusters_flow(x, params):
    """
    Single bifurcation followed by bifurcation of each cluster
    """

    if x[0] < 2:
        v = np.zeros(x.shape)
        v[0] = params.x0_speed
        v[1] = -x[1]**3 + x[0]*x[1]
        v[2] = -x[2]
        return v
    elif x[0] < 4:
        v = np.zeros(x.shape)
        v[0] = params.x0_speed

        stationary_spots = np.array([(x[0] - 1)**(-1),(2 - (x[0]-1)**(-1)) , 1])*1.25
        stationary_spots = np.concatenate([stationary_spots, [0], -stationary_spots])
        v[1] = (-1)**(np.sum(x[1] > stationary_spots))
        

        #add split in 3rd dimension
        v[2] = v[0]*2*np.sign(x[1])
        return v
    else:
        v = np.zeros(x.shape)
        v[0] = params.x0_speed

        
        stationary_spots = np.array([-0.1, 2/3 , 1.1])*np.sqrt(2)
        if x[2] < 0:
            stationary_spots = -stationary_spots

        v[1] = -np.prod(x[1] - stationary_spots)/2

        v[2] = -np.prod(x[2] - np.array([-3, 0, 3]))/1000
        return v


def partial_convergent_flow(x, params):
    """
    Single bifurcation followed by bifurcation of each cluster,
    where two of the new clusters subsequently merge
    """

    if x[0] < 2:
        # First bifurcation
        v = np.zeros(x.shape)
        v[0] = params.x0_speed
        v[1] = -x[1]**3 + x[0]*x[1]
        return v
    elif x[0] < 4:
        # Second bifurcation
        v = np.zeros(x.shape)
        v[0] = params.x0_speed

        stationary_spots = np.array([(x[0] - 1)**(-1),(2 - (x[0]-1)**(-1)) , 1])*1.25
        stationary_spots = np.concatenate([stationary_spots, [0], -stationary_spots])
        v[1] = (-1)**(np.sum(x[1] > stationary_spots))


        return v
    else:
        # Convergence
        v = np.zeros(x.shape)
        v[0] = params.x0_speed


        stationary_spots = np.array([-1.1, -2/3, 0, 2/3, 1.1])*np.sqrt(2)


        v[1] = -np.prod(x[1] - stationary_spots)/(2)
        return v




def convergent_flow(x, params):
    """
    Single bifurcation followed by convergence of the two clusters
    """

    v = np.zeros(x.shape)
    v[0] = params.x0_speed

    if x[0] < 2:
        # First bifurcation
        v[1] = -x[1]**3 + x[0]*x[1]
        return v
    else:
        # Convergence
        v[1] = -0.6*x[1]*params.x0_speed
        return v







def vector_field(x, params):
    """
    Selects a vector field and returns its value at x
    """

    if params.flow_type == "bifurcation":
        return single_bifurcation_flow(x)
    elif params.flow_type == "mismatched_clusters":
        return mismatched_clusters_flow(x, params)
    elif params.flow_type == "partial_convergent":
        return partial_convergent_flow(x, params)
    elif params.flow_type == "convergent":
        return convergent_flow(x, params)
    else:
        warnings.warn("Unknown vector field option. Vector field set to zero.")
        return np.zeros(x.shape)








def center(barcode, params):
    """
    Returns the center of the distribution p(x0|barcode)
    """
    c = np.zeros(params.num_genes)
    l = min(params.num_genes, params.barcode_length)
    c[0:l] = barcode[0:l]
    return c


def sample_barcode(params):
    """
    Samples an initial barcode
    """
    if params.back_mutations:
        barcode = np.random.randint(0,params.alphabet_size,params.barcode_length)
    else:
        barcode = np.zeros(params.barcode_length)
    return barcode

def sample_x0(barcode, params):
    """
    Samples the initial position in gene expression space
    """
    return center(barcode, params) + params.initial_distribution_std*np.random.randn(params.num_genes)

def sample_cell(params):
    """
    Samples an initial cell
    """
    barcode = sample_barcode(params)
    x0 = sample_x0(barcode, params)
    return Cell(x0, barcode)


def evolve_x(initial_x, time, params):
    """
    Returns a sample from Langevin dynamics following potential_gradient
    """
    current_x = initial_x
    if params.flow_type != None:
        current_time = 0
        #print("Evolving x for time %f\n"% time)
        while current_time < time:
            current_x = (current_x 
                         + params.timestep*vector_field(current_x, params)
                         + np.sqrt(params.timestep)*np.dot(params.diffusion_matrix,
                                                           np.random.randn(initial_x.size)))
            current_time = current_time + params.timestep
        #print("Finished evolving x\n")
    return current_x


def sample_division_time(params):
    """
    Samples the time until a cell divides
    """
    t = -1
    for _ in range(10): # try 10 times to get a positive time
        if params.division_time_distribution == 'normal':
            t = np.random.normal(params.mean_division_time, params.division_time_std)
        elif params.division_time_distribution == 'exponential':
            t = np.random.exponential(params.mean_division_time)
        else:
            t = params.mean_division_time
        if t > 0:
            break
    return t
    
#

def evolve_b(initial_barcode, time, params):
    """
    Returns the new barcode after mutations have occurred for some time
    """
    # Setting the seed for reproducibility, specifically to make the
    # process consistent with different time intervals.
    # If the random seed has been set before evolve_b is called, then
    # we want the mutations that occur in time t1 to be a subset of the
    # mutations that occur in time t2 if t2 > t1.

    mutation_seed = np.random.randint(2**32)

    rate = time*params.mutation_rate
    if (rate > 10) & params.enforce_barcode_reproducibility:
        warnings.warn("Large number of mutations expected on one edge. " +
                      "Falling back on slow non-numpy Poisson sampling.")
        num_mutations = reproducible_poisson(rate)
    else:
        num_mutations = np.random.poisson(rate)

    # need to reset the seed after the Poisson sample because the state of the
    # RNG after Poisson sampling depends on the parameter of the Poisson distribution
    np.random.seed(mutation_seed)
    for m in range(num_mutations):
        mutate_barcode(initial_barcode, params)
    return initial_barcode

def mutate_barcode(barcode, params):
    """
    Randomly changes one entry of the barcode
    """
    changed_entry = np.random.randint(0, len(barcode)) # not modeling different mutation rates across cut sites
    if params.back_mutations | (barcode[changed_entry] == 0):
        # if no back mutations, only change unmutated sites
        barcode[changed_entry] = np.random.choice(range(params.alphabet_size), p = params.mutation_likelihoods)
    return barcode

def evolve_cell(initial_cell, time, params):
    """
    Returns a new cell after both barcode and x have evolved for some time
    """
    np.random.seed(initial_cell.seed) # allows reproducibility of individual trajectories
    new_b = evolve_b(initial_cell.barcode, time, params)

    np.random.seed(initial_cell.seed + 1)
    new_x = evolve_x(initial_cell.x, time, params)

    if params.keep_cell_seeds:
        return Cell(new_x, new_b, seed = initial_cell.seed)
    else:
        return Cell(new_x, new_b)


def mask_barcode(barcode, p):
    """
    Replaces a subset of the entries of barcode with -1 to simulate missing data

    Entries are masked independently with probability p 

    Also works for an array of barcodes
    """
    mask = np.random.rand(*barcode.shape) < p
    barcode[mask] = -1
    return barcode







def sample_descendants(initial_cell, time, params, target_num_cells = None):
    """
    Samples the descendants of an initial cell
    """
    assert(isinstance(initial_cell, Cell))

    if target_num_cells == None:
        target_num_cells = params.target_num_cells

    next_division_time = sample_division_time(params)
    if (next_division_time > time) | (target_num_cells == 1):
        #print("In final division stage\n")
        if params.keep_tree:
            return[(evolve_cell(initial_cell, time, params), time)]
        else:
            return [evolve_cell(initial_cell, time, params)]
    else:
        time_remaining = time - next_division_time
        target_num_cells_1, target_num_cells_2 = split_targets_between_daughters(time_remaining, target_num_cells, params)

        while (target_num_cells_1 == 0) | (target_num_cells_2 == 0):
            # wait until we get a division where both daughters have observed descendants
            t = sample_division_time(params)
            if t > time_remaining:
                # In rare cases, you can have:
                # 1) time_remaining greater than mean_division_time
                # 2) split_targets assigning 0 cells to one branch so we skip to the next division and
                # 3) sample_division_time greater than time_remaining
                # leading to negative time_remaining and an error
                #
                # If mean_division_time >> division_time_std, this will only happen at the last division
                # which is the case this handles
                if target_num_cells == 2:
                    # In this case, since the branch we chose doesn't divide before the end of the simulation,
                    # we can assign one of its samples to the other branch
                    target_num_cells_1 = 1
                    target_num_cells_2 = 1
                    break
                else:
                    raise(ValueError)
                
            else:
                next_division_time = next_division_time + t
                time_remaining = time - next_division_time
                target_num_cells_1, target_num_cells_2 = split_targets_between_daughters(time_remaining, target_num_cells, params)

        cell_at_division = evolve_cell(initial_cell, next_division_time, params)
        assert(cell_at_division.seed >= 0)
        daughter_1 = cell_at_division.deepcopy()
        daughter_2 = cell_at_division.deepcopy()

        daughter_1.reset_seed()
        daughter_2.reset_seed()
        
        if params.keep_tree:
            # store as list of lists that keeps tree structure, ancestors, and ancestor times
            return  [sample_descendants(daughter_1, time_remaining, params, target_num_cells_1), 
                     sample_descendants(daughter_2, time_remaining, params, target_num_cells_2),
                     (cell_at_division, next_division_time)]
        else:
            # store all cells as a flat list, ignoring ancestors
            return (sample_descendants(daughter_1, time_remaining, params, target_num_cells_1) +
                    sample_descendants(daughter_2, time_remaining, params, target_num_cells_2) )
    return


def split_targets_between_daughters(time_remaining, target_num_cells, params):
    """
    Given a target number of cells to sample, divides the samples between daughters
    assuming both have the expected number of descendants at the sampling time
    """
    num_future_generations = np.floor(time_remaining/params.mean_division_time)
    num_descendants = 2**num_future_generations
    
    if target_num_cells > 2*num_descendants:
        # we expect to sample all the cells
        target_num_cells_1 = np.floor(target_num_cells/2)
    else:
        target_num_cells_1 = np.random.hypergeometric(num_descendants, num_descendants, target_num_cells)

    if target_num_cells == np.inf:
        target_num_cells_2 = np.inf
    else:
        target_num_cells_2 = target_num_cells - target_num_cells_1

    return target_num_cells_1, target_num_cells_2



def sample_population_descendants(pop, time, params):
    """
    Samples the descendants of each cell in a population
    pop: list of (expression, barcode) tuples
    """
    sampled_population = []
    num_descendants = np.zeros((len(pop))) # the number of descendants of each cell
    for cell in range(len(pop)):
        descendants = sample_descendants(pop[cell], time, params)
        num_descendants[cell] = len(descendants)
        sampled_population = sampled_population + descendants

    return sampled_population, num_descendants

def flatten_list_of_lists(tree_data):
    """
    Converts a dataset of cells with their ancestral tree structure to a list of cells
    (with ancestor and time information dropped)
    """
    assert isinstance(tree_data, list)
    if len(tree_data) == 1:
        assert(isinstance(tree_data[0][0], Cell))
        return [tree_data[0][0]]
    else:
        assert len(tree_data) == 3 # assuming this is a binary tree
        # the third entry of tree_data should be a (cell, time) tuple
        assert isinstance(tree_data[2][0], Cell)
        return flatten_list_of_lists(tree_data[0]) + flatten_list_of_lists(tree_data[1])
    return

def convert_data_to_arrays(data):
    """
    Converts a list of cells to two ndarrays,
    one for expression and one for barcodes
    """
    expressions = np.array([cell.x for cell in data])
    barcodes = np.array([cell.barcode for cell in data])
    return expressions, barcodes



def sample_pop(num_initial_cells, time, params):
    """
    Samples a population after some intervening time

    num_initial_cells:                          Number of cells in the population at time 0
    time:                                       Time when population is measured
    params:                                     Simulation parameters
    """

    if num_initial_cells == 0:
        return np.zeros((0, params.num_genes)), np.zeros((0, params.barcode_length)), np.zeros((0)), []
    
    initial_population = []
    for cell in range(num_initial_cells):
        initial_population = initial_population + [sample_cell(params)]

    sampled_population, num_descendants = sample_population_descendants(initial_population, time, params)

    expressions, barcodes = convert_data_to_arrays(sampled_population)
    return (expressions, barcodes, num_descendants, initial_population)




def subsample_list(sample, target_num_cells):
    """
    Randomly samples target_num_cells from the sample
    
    If there are fewer than target_num_cells in the sample,
    returns the whole sample
    """

    if target_num_cells > len(sample):
        return sample
    else:
        # not using permutation so the order of elements in sample is preserved
        r = np.random.rand(len(sample))
        sorted_indices = np.argsort(r)
        min_dropped_r = r[sorted_indices[target_num_cells]]
        return sample[r < min_dropped_r]





def subsample_pop(sample, target_num_cells, params, num_cells = None):
    """
    Randomly samples target_num_cells from the sample. Subsampling during the simulation by
    setting params.target_num_cells is a more efficient approximation of this.
    
    If there are fewer than target_num_cells in the sample,
    returns the whole sample

    sample should be either:

    - a list of cells, if params.keep_tree is False
    - nested lists of lists of cells encoding the tree structure, if params.keep_tree is True

    (i.e., it should match the output of sample_descendants with the same params)
    """

    if target_num_cells == 0:
        return []
    elif params.keep_tree:
        if num_cells == None:
            sample_list = flatten_list_of_lists(sample)
            num_cells = len(sample_list)

        if num_cells <= target_num_cells:
            return sample
        else:
            daughter_1_subtree = sample[0]
            daughter_2_subtree = sample[1]
            
            # TODO: there is a lot of redundant flattening happening. Should be
            #       redone if it is slow
            num_cells_1 = len(flatten_list_of_lists(daughter_1_subtree))
            num_cells_2 = len(flatten_list_of_lists(daughter_2_subtree))

            target_num_cells_1 = np.random.hypergeometric(num_cells_1, num_cells_2, target_num_cells)
            target_num_cells_2 = target_num_cells - target_num_cells_1

            # If one subtree does not get sampled, return only the other subtree subsampled
            # with its root 'time_to_parent' adjusted
            if target_num_cells_1 == 0:
                daughter_2_subtree[-1] = (daughter_2_subtree[-1][0], daughter_2_subtree[-1][1] + sample[2][1])
                return subsample_pop(daughter_2_subtree, target_num_cells_2, params, num_cells = num_cells_2)
            elif target_num_cells_2 == 0:
                daughter_1_subtree[-1] = (daughter_1_subtree[-1][0], daughter_1_subtree[-1][1] + sample[2][1])
                return subsample_pop(daughter_1_subtree, target_num_cells_1, params, num_cells = num_cells_1)
            else:
                return [subsample_pop(daughter_1_subtree, target_num_cells_1, params, num_cells = num_cells_1),
                        subsample_pop(daughter_2_subtree, target_num_cells_2, params, num_cells = num_cells_2),
                        sample[2]]

            

    else:
        return subsample_list(sample, target_num_cells)

    










def reproducible_poisson(rate):
    """
    Samples a single Poisson random variable, in a way
    that is reproducible, i.e. after
    
    np.random.seed(s)
    a = divisible_poisson(r1)
    np.random.seed(s)
    b = divisible_poisson(r2)
    
    with r1 > r2, b ~ binomial(n = a, p = r2/r1)


    This is the standard numpy Poisson sampling algorithm for rate <= 10.
    
    Note that this is relatively slow, running in O(rate) time.
    """
    T = np.exp(-rate)
    k = -1
    t = 1
    while t >= T:
        t = t*np.random.rand()
        k = k+1
        
    return k



# Evaluating fitted couplings
# Also includes some utility functinos for manipulating and plotting couplings


import numpy as np
import ot
import warnings
import matplotlib.pyplot as plt

def l2_difference(coupling_1, coupling_2):
    return np.linalg.norm(coupling_1 - coupling_2, 'fro')

def scaled_l2_difference(coupling_1, coupling_2):
    return l2_difference(coupling_1, coupling_2)/np.linalg.norm(coupling_2, 'fro')


def pairwise_squared_distances(data):
    """
    Returns the pairwise squared distances between rows of the data matrix
    """
    n = data.shape[0]
    distances = np.zeros((n,n))
    for i in range(n):
        for j in range(i+1,n):
            d = data[i,:]-data[j,:]
            distances[i,j] = d@d
            distances[j,i] = distances[i,j]

    return distances
            


def coupling_to_coupling_cost_matrix(source, target):
    """
    Returns the (n_source*n_target)*(n_source*n_target) cost matrix
    for a W2 distance between two couplings of source and target

    Source and target here are just expression samples, without barcodes
    """
    n_source = source.shape[0]
    n_target = target.shape[0]

    cost_matrix = np.zeros((n_source*n_target, n_source*n_target))

    source_distances = pairwise_squared_distances(source)
    target_distances = pairwise_squared_distances(target)

    #UPDATE: vectorize
    for source_1 in range(n_source):
        for target_1 in range(n_target):
            index_1 = source_1*n_target + target_1
            for source_2 in range(source_1, n_source):
                for target_2 in range(n_target):
                    index_2 = source_2*n_target + target_2
                    
                    cost_matrix[index_1, index_2] = (source_distances[source_1, source_2]
                                                   + target_distances[target_1, target_2])
                    cost_matrix[index_2, index_1] = cost_matrix[index_1, index_2]
            
    return cost_matrix


def coupling_W2(coupling_1, coupling_2, source, target, epsilon):
    """
    Returns the entropically-regularized W2 distance between two couplings
    """
    cost_matrix = coupling_to_coupling_cost_matrix(source, target)
    return ot.sinkhorn2(coupling_1.flatten(), coupling_2.flatten(), cost_matrix, epsilon)








def squeeze_coupling_by_late_cluster(c, index):
    c1 = np.sum(c[:,  index], 1)
    c2 = np.sum(c[:, ~index], 1)
    return np.array([c1, c2]).T

def squeeze_coupling(c, row_cluster_labels = None, column_cluster_labels = None):
    if row_cluster_labels is None:
        row_cluster_labels = range(c.shape[0])
    if column_cluster_labels is None:
        column_cluster_labels = range(c.shape[1])
    row_clusters = np.unique(row_cluster_labels)
    num_row_clusters = len(row_clusters)
    column_clusters = np.unique(column_cluster_labels)
    num_column_clusters = len(column_clusters)
    
    squeezed_c = np.zeros([num_row_clusters, num_column_clusters])
    # if you have lots of clusters find a way to do this loop in numpy
    for i in range(num_row_clusters):
        for j in range(num_column_clusters):
            squeezed_c[i,j] = np.sum(c[np.ix_(row_cluster_labels == row_clusters[i],
                                              column_cluster_labels == column_clusters[j])])
    return squeezed_c

def tv(coupling1, coupling2):
    return np.linalg.norm((coupling1 - coupling2).flatten(), 1)/2

def normalize_columns(coupling):
    s = np.sum(coupling, 0)
    n = coupling.shape[1]
    return coupling*(s**(-1))/n



def expand_coupling_independent(c, true_coupling):
    s = np.sum(true_coupling, 1)
    conditional_coupling = true_coupling.T*(s**(-1))
    return conditional_coupling@c

def expand_coupling(c, true_coupling, distances, matched_dim = 0, max_dims_used = np.inf, xs_used = None):
    """
    Parameters
    ----------
    c : ndarray, shape (nx, ny) if matched_dim == 0, (ny, nx) if matched_dim == 1
        Coupling between source x and target y
    true_coupling : ndarray, shape (nx, nz) if matched_dim == 0, (nz, nx) if matched_dim == 1
        Reference coupling between x and z
    distances : ndarray, shape (nz, ny)
        Pairwise distances between z and y
    matched_dim : int
        Dimension in which c and true coupling
    max_dims_used: int or np.inf
        Set a finite value here to do an approximate calculation based on min(nx, max_dims_used) elements of x
    xs_used : list or None
        Indices of matched_dim to use in approximate calculation. If None and max_dims_used<nx, indices are randomly selected.
    
    Returns
    -------
    expanded_coupling : ndarray, shape same as true_couplings
        Optimal coupling between z and y consistent with the coupling c
    """
    if matched_dim == 1:
        c = c.T
        true_coupling = true_coupling.T
        
    expanded_coupling = np.zeros([true_coupling.shape[1], c.shape[1]])
    
    n_x = c.shape[0]
    x_marginal = np.sum(c, 1)
    if max_dims_used < x_marginal.size and xs_used is None:
        xs_used = np.random.choice(range(x_marginal.size), size = max_dims_used, p = x_marginal, replace = False)

    if not xs_used is None:
        xs_not_used = [i for i in range(x_marginal.size) if not i in xs_used]
        x_marginal[xs_not_used] = 0
        x_marginal = x_marginal/np.sum(x_marginal)
    for i in range(n_x):
        p_true = true_coupling[i, :]/np.sum(true_coupling[i, :])
        p_c = c[i, :]/np.sum(c[i, :])
        if x_marginal[i] > 0:
            expanded_coupling = expanded_coupling + ot.emd(p_true, p_c, distances)*x_marginal[i]
    
    if abs(np.sum(expanded_coupling) - 1) > 10**(-8):
        warnings.warn(("Expanded coupling not computed correctly. " +
                       "Check for infeasibility in row and column OT calculations. Updating the python optimal transport package may help (https://github.com/PythonOT/POT/issues/93).\n" +
                       "If total mass - 1 is small, this may not significantly affect downstream results.\n" +
                       "Total mass - 1: " +
                       str(np.sum(expanded_coupling) -1)))

    if matched_dim == 1:
        return expanded_coupling.T
    else:
        return expanded_coupling



def sample_indices_from_coupling(c, num_samples=None, return_all = False, thr = 10**(-6)):
    """
    Generates [row, column] samples from the coupling c

    If return_all is True, then returns all indices with coupling values above the threshold
    """

    if return_all:
        return [[i, j] for i in range(c.shape[0]) for j in range(c.shape[1]) if c[i,j] > thr]
    else:
        linear_samples = np.random.choice(range(c.size), size = num_samples, p = c.flat)
        
        if num_samples in [1, None]:
            linear_samples = [linear_samples]
            
        n_cols = c.shape[1]
        return np.array([[s//n_cols  , s%n_cols ] for s in linear_samples])



def sample_coordinates_from_coupling(c, row_points, column_points, num_samples=None, return_all = False, thr = 10**(-6)):
    """
    Generates [x, y] samples from the coupling c.

    If return_all is True, returns [x,y] coordinates of every pair with coupling value >thr
    """

    index_samples = sample_indices_from_coupling(c, num_samples = num_samples, return_all = return_all, thr = thr)

    return np.array([ [row_points[s[0], :], column_points[s[1],:]] for s in index_samples])


def sample_interpolant(coupling, row_points, column_points, t=0.5, num_samples=None, return_all=False, thr=10**(-6)):
    """
    Samples from the interpolated distribution implied by the coupling

    If return_all is True, returns the interpolants between every pair with coupling value >thr. This is the exact interpolant
    distribution if and only if all nonzero values of the coupling are identical and >thr.
    """

    # Sample start and end points
    coordinates = sample_coordinates_from_coupling(coupling,
                                                   row_points,
                                                   column_points,
                                                   num_samples = num_samples,
                                                   return_all = return_all,
                                                   thr = thr)

    # Average start and end points, weighted by t
    return np.array([(1-t)*pair[0] + t*pair[1] for pair in coordinates])



def plot2D_samples_mat(xs, xt, G, thr=1e-8, alpha_scale = 1, **kwargs):
    """ Plot matrix M  in 2D with  lines using alpha values

    Plot lines between source and target 2D samples with a color
    proportional to the value of the matrix G between samples.

    Copied function from PythonOT and added alpha_scale parameter


    Parameters
    ----------
    xs : ndarray, shape (ns,2)
        Source samples positions
    b : ndarray, shape (nt,2)
        Target samples positions
    G : ndarray, shape (na,nb)
        OT matrix
    thr : float, optional
        threshold above which the line is drawn
    **kwargs : dict
        parameters given to the plot functions (default color is black if
        nothing given)
    """
    if ('color' not in kwargs) and ('c' not in kwargs):
        kwargs['color'] = 'k'
    mx = G.max()
    for i in range(xs.shape[0]):
        for j in range(xt.shape[0]):
            if G[i, j] / mx > thr:
                plt.plot([xs[i, 0], xt[j, 0]], [xs[i, 1], xt[j, 1]],
                        alpha=alpha_scale*G[i, j] / mx, **kwargs)


def print_metrics(couplings, cost_func, cost_func_name, log = False):
    """
    Prints cost_func evaluated for each coupling in the dictionary couplings
    """
    
    l = max([len(c) for c in couplings.keys()])
    print(cost_func_name)
    for c in couplings.keys():
        loss = cost_func(couplings[c])
        if log:
            loss = np.log(loss)
        print(c.ljust(l), ": ", "{:.3f}".format(loss))
    print("\n")
    return


def plot_metrics(couplings, cost_func, cost_func_name, epsilons, log = False, points=False, scale=1.0, label_font_size=18, tick_font_size=12):
    """
    Plots cost_func evaluated as a function of epsilon
    """
    zero_offset = epsilons[0]/2
    all_ys = []
    if "lineageOT" in couplings.keys():
        ys = np.array([cost_func(c) for c in [couplings['lineage entropic rna ' + str(e)] for e in epsilons]])
        plt.plot(epsilons, ys/scale, label = "LineageOT, true tree")
        if points:
            plt.scatter([zero_offset], [cost_func(couplings["lineageOT"])/scale])
        all_ys.append(ys)
    if "OT" in couplings.keys():
        ys = np.array([cost_func(c) for c in [couplings['entropic rna ' + str(e)] for e in epsilons]])
        plt.plot(epsilons, ys/scale, label = "Entropic OT")
        if points:
            plt.scatter([zero_offset], [cost_func(couplings["OT"])/scale])
        all_ys.append(ys)
    if "lineageOT, fitted" in couplings.keys():
        ys = np.array([cost_func(c) for c in [couplings['fitted lineage rna ' + str(e)] for e in epsilons]])
        plt.plot(epsilons, ys/scale, label = "LineageOT, fitted tree")
        if points:
            plt.scatter([zero_offset], [cost_func(couplings["lineageOT, fitted"])/scale])
        all_ys.append(ys)

    plt.ylabel(cost_func_name, fontsize=label_font_size)
    plt.xlabel("Entropy parameter", fontsize=label_font_size)
    plt.xscale("log")

    plt.xticks(fontsize=tick_font_size)
    plt.yticks(fontsize=tick_font_size)

    if points:
        plt.xlim([0.9*zero_offset, epsilons[-1]])
    else:
        plt.xlim([epsilons[0], epsilons[-1]])

    ylims = plt.ylim([0, None])
    # upper limit should be at least 1
    plt.ylim([0, max(ylims[1], 1)])

    plt.legend(fontsize=tick_font_size)
    return all_ys
